from __future__ import absolute_import, division, print_function
from toolz import memoize
from .drop import drop
from .utils import copydoc


class _Temp(object):
    """ Temporary version of persistent storage

    Calls ``drop`` on object at garbage collection

    This is a parametrized type, so call it on types to make new types

    >>> from odo import Temp, CSV
    >>> csv = Temp(CSV)('/tmp/myfile.csv', delimiter=',')
    """
    def __del__(self):
        drop(self)


@memoize
@copydoc(_Temp)
def Temp(cls):
    """ Parametrized Chunks Class """
    return type('Temp(%s)' % cls.__name__, (_Temp, cls), {'persistent_type': cls})



from odo.utils import ext, iter_except, keywords, gentemp, records_to_tuples


def test_ext():
    assert ext('a.py') == 'py'
    assert ext('apy') == ''
    assert ext('myfile.csv.gz') == 'gz'


def test_iter_except():
    s = set([1, 2, 3])

    def iterate():
        yield 1
        yield 2
        yield 4

    elems = iterate()
    inset = iter_except(lambda: next(elems) in s, StopIteration)
    assert list(inset) == [True, True, False]


def test_keywords():
    assert keywords(keywords) == ["func"]


def test_gentemp():
    i, _, data = next(gentemp([[1, 2], [3, 4]], start=1))
    assert i == 1 and data == [1, 2]


def test_records_to_tuples():
    seq = [{'i': 1, 'f': 1.0}, {'i': 2, 'f': 2.0}]
    lst = list(records_to_tuples('var * {i: int, f: float64}', seq))
    assert lst == [(1, 1.0), (2, 2.0)]


def test_records_to_tuples_mismatch_passthrough():
    assert records_to_tuples('var * int', 'dummy') == 'dummy'



from __future__ import absolute_import, division, print_function

import datetime
import os
import gzip
import os
import json

from contextlib import contextmanager

import numpy as np
from odo.backends.json import json_dumps
from odo.utils import tmpfile, ignoring
from odo import odo, discover, JSONLines, resource, JSON, convert, append, drop
from odo.temp import Temp, _Temp

from datashape import dshape


@contextmanager
def json_file(data):
    with tmpfile('.json') as fn:
        with open(fn, 'w') as f:
            json.dump(data, f, default=json_dumps)

        yield fn


@contextmanager
def jsonlines_file(data):
    with tmpfile('.json') as fn:
        with open(fn, 'w') as f:
            for item in data:
                json.dump(item, f, default=json_dumps)
                f.write('\n')

        yield fn


dat = [{'name': 'Alice', 'amount': 100},
       {'name': 'Bob', 'amount': 200}]


def test_discover_json():
    with json_file(dat) as fn:
        j = JSON(fn)
        assert discover(j) == discover(dat)


def test_discover_jsonlines():
    with jsonlines_file(dat) as fn:
        j = JSONLines(fn)
        assert discover(j) == discover(dat)


def test_discover_json_only_includes_datetimes_not_dates():
    data = [{'name': 'Alice', 'dt': datetime.date(2002, 2, 2)},
            {'name': 'Bob',   'dt': datetime.date(2000, 1, 1)}]
    with json_file(data) as fn:
        j = JSON(fn)
        assert discover(j) == dshape('2 * {dt: datetime, name: string }')


def test_resource():
    with tmpfile('json') as fn:
        assert isinstance(resource('jsonlines://' + fn), JSONLines)
        assert isinstance(resource('json://' + fn), JSON)

        assert isinstance(resource(fn, expected_dshape=dshape('var * {a: int}')),
                          JSONLines)


def test_resource_guessing():
    with json_file(dat) as fn:
        assert isinstance(resource(fn), JSON)

    with jsonlines_file(dat) as fn:
        assert isinstance(resource(fn), JSONLines)


def test_append_jsonlines():
    with tmpfile('json') as fn:
        j = JSONLines(fn)
        append(j, dat)
        with open(j.path) as f:
            lines = f.readlines()
        assert len(lines) == 2
        assert 'Alice' in lines[0]
        assert 'Bob' in lines[1]


def test_append_json():
    with tmpfile('json') as fn:
        j = JSON(fn)
        append(j, dat)
        with open(j.path) as f:
            lines = f.readlines()
        assert len(lines) == 1
        assert 'Alice' in lines[0]
        assert 'Bob' in lines[0]


def test_convert_json_list():
    with json_file(dat) as fn:
        j = JSON(fn)
        assert convert(list, j) == dat


def test_convert_jsonlines():
    with jsonlines_file(dat) as fn:
        j = JSONLines(fn)
        assert convert(list, j) == dat


def test_tuples_to_json():
    ds = dshape('var * {a: int, b: int}')
    with tmpfile('json') as fn:
        j = JSON(fn)

        append(j, [(1, 2), (10, 20)], dshape=ds)
        with open(fn) as f:
            assert '"a": 1' in f.read()

    with tmpfile('json') as fn:
        j = JSONLines(fn)

        append(j, [(1, 2), (10, 20)], dshape=ds)
        with open(fn) as f:
            assert '"a": 1' in f.read()


def test_datetimes():
    from odo import into
    import numpy as np
    data = [{'a': 1, 'dt': datetime.datetime(2001, 1, 1)},
            {'a': 2, 'dt': datetime.datetime(2002, 2, 2)}]
    with tmpfile('json') as fn:
        j = JSONLines(fn)
        append(j, data)

        assert str(into(np.ndarray, j)) == str(into(np.ndarray, data))


def test_json_encoder():
    result = json.dumps([1, datetime.datetime(2000, 1, 1, 12, 30, 0)],
                        default=json_dumps)
    assert result == '[1, "2000-01-01T12:30:00Z"]'
    assert json.loads(result) == [1, "2000-01-01T12:30:00Z"]


def test_empty_line():
    text = '{"a": 1}\n{"a": 2}\n\n'  # extra endline
    with tmpfile('.json') as fn:
        with open(fn, 'w') as f:
            f.write(text)
        j = JSONLines(fn)
        assert len(convert(list, j)) == 2


def test_multiple_jsonlines():
    a, b = '_test_a1.json', '_test_a2.json'
    try:
        with ignoring(OSError):
            os.remove(a)
        with ignoring(OSError):
            os.remove(b)
        with open(a, 'w') as f:
            json.dump(dat, f)
        with open(b'_test_a2.json', 'w') as f:
            json.dump(dat, f)
        r = resource('_test_a*.json')
        result = convert(list, r)
        assert len(result) == len(dat) * 2
    finally:
        with ignoring(OSError):
            os.remove(a)
        with ignoring(OSError):
            os.remove(b)


def test_read_gzip_lines():
    with tmpfile('json.gz') as fn:
        f = gzip.open(fn, 'wb')
        for item in dat:
            s = json.dumps(item).encode('utf-8')
            f.write(s)
            f.write(b'\n')
        f.close()
        js = JSONLines(fn)
        assert convert(list, js) == dat


def test_read_gzip():
    with tmpfile('json.gz') as fn:
        f = gzip.open(fn, 'wb')
        s = json.dumps(dat).encode('utf-8')
        f.write(s)
        f.close()
        js = JSON(fn)
        assert convert(list, js) == dat


def test_write_gzip_lines():
    with tmpfile('json.gz') as fn:
        j = JSONLines(fn)
        append(j, dat)

        f = gzip.open(fn)
        line = next(f)
        f.close()
        assert line.decode('utf-8').strip() == str(json.dumps(dat[0]))


def test_write_gzip():
    with tmpfile('json.gz') as fn:
        j = JSON(fn)
        append(j, dat)

        f = gzip.open(fn)
        text = f.read()
        f.close()
        assert text.decode('utf-8').strip() == str(json.dumps(dat))
        assert isinstance(resource(fn), (JSON, JSONLines))


def test_resource_gzip():
    with tmpfile('json.gz') as fn:
        assert isinstance(resource(fn), (JSON, JSONLines))
        assert isinstance(resource('json://' + fn), (JSON, JSONLines))
        assert isinstance(resource('jsonlines://' + fn), (JSON, JSONLines))

    with tmpfile('jsonlines.gz'):
        assert isinstance(resource('jsonlines://' + fn), (JSON, JSONLines))


def test_convert_to_temp_json():
    js = convert(Temp(JSON), [1, 2, 3])
    assert isinstance(js, JSON)
    assert isinstance(js, _Temp)

    assert convert(list, js) == [1, 2, 3]


def test_drop():
    with tmpfile('json') as fn:
        js = JSON(fn)
        append(js, [1, 2, 3])

        assert os.path.exists(fn)
        drop(js)
        assert not os.path.exists(fn)


def test_missing_to_csv():
    data = [dict(a=1, b=2), dict(a=2, c=4)]
    with tmpfile('.json') as fn:
        js = JSON(fn)
        js = odo(data, js)

        with tmpfile('.csv') as csvf:
            csv = odo(js, csvf)
            with open(csv.path, 'rt') as f:
                result = f.read()

    expected = 'a,b,c\n1,2.0,\n2,,4.0\n'
    assert result == expected



from __future__ import absolute_import, division, print_function

import pytest
pytest.importorskip('bcolz')

from odo.backends.bcolz import (append, convert, ctable, carray, resource,
                                 discover, drop)
from odo.chunks import chunks
from odo import append, convert, discover, into
import numpy as np
from odo.utils import tmpfile, ignoring, filetext
from contextlib import contextmanager
import shutil
import os
import uuid


@contextmanager
def tmpbcolz(*args, **kwargs):
    fn = '.%s.bcolz' % str(uuid.uuid1())
    r = resource(fn, *args, **kwargs)

    try:
        yield r
    finally:
        with ignoring(Exception):
            r.flush()
        if os.path.exists(fn):
            shutil.rmtree(fn)



def eq(a, b):
    c = a == b
    if isinstance(c, np.ndarray):
        c = c.all()
    return c


a = carray([1, 2, 3, 4])
x = np.array([1, 2])


def test_discover():
    assert discover(a) == discover(a[:])


def test_convert():
    assert isinstance(convert(carray, np.ones([1, 2, 3])), carray)
    b = carray([1, 2, 3])
    assert isinstance(convert(np.ndarray, b), np.ndarray)


def test_chunks():
    c = convert(chunks(np.ndarray), a, chunksize=2)
    assert isinstance(c, chunks(np.ndarray))
    assert len(list(c)) == 2
    assert eq(list(c)[1], [3, 4])

    assert eq(convert(np.ndarray, c), a[:])


def test_append_chunks():
    b = carray(x)
    append(b, chunks(np.ndarray)([x, x]))
    assert len(b) == len(x) * 3


def test_append_other():
    b = carray(x)
    append(b, convert(list, x))
    assert len(b) == 2 * len(x)


def test_resource_ctable():
    with tmpbcolz(dshape='var * {name: string[5, "ascii"], balance: int32}') as r:
        assert isinstance(r, ctable)
        assert r.dtype == [('name', 'S5'), ('balance', 'i4')]


def get_expectedlen(x):
    reader = getattr(x, '_read_meta', getattr(x, 'read_meta', None))
    assert reader is not None
    shape, cparams, dtype, _, expectedlen, _, chunklen = reader()
    return expectedlen


def test_resource_ctable_overrides_expectedlen():
    with tmpbcolz(dshape='100 * {name: string[5, "ascii"], balance: int32}',
                  expectedlen=200) as r:
        assert isinstance(r, ctable)
        assert r.dtype == [('name', 'S5'), ('balance', 'i4')]
        assert all(get_expectedlen(r[c]) == 200 for c in r.names)


def test_resource_ctable_correctly_infers_length():
    with tmpbcolz(dshape='100 * {name: string[5, "ascii"], balance: int32}') as r:
        assert isinstance(r, ctable)
        assert r.dtype == [('name', 'S5'), ('balance', 'i4')]
        assert all(get_expectedlen(r[c]) == 100 for c in r.names)


def test_resource_carray():
    with tmpbcolz(dshape='var * int32') as r:
        assert isinstance(r, carray)
        assert r.dtype == 'i4'
        assert r.shape == (0,)


def test_resource_existing_carray():
    with tmpbcolz(dshape='var * int32') as r:
        append(r, [1, 2, 3])
        r.flush()
        newr = resource(r.rootdir)
        assert isinstance(newr, carray)


def test_resource_carray_overrides_expectedlen():
    with tmpbcolz(dshape='100 * int32', expectedlen=200) as r:
        assert isinstance(r, carray)
        assert r.dtype == 'i4'
        assert r.shape == (100,)
        assert get_expectedlen(r) == 200


def test_resource_ctable_correctly_infers_length():
    with tmpbcolz(dshape='100 * int32') as r:
        assert isinstance(r, carray)
        assert r.dtype == 'i4'
        assert get_expectedlen(r) == 100


def test_into_respects_expected_len_during_append():
    with tmpfile('.bcolz') as fn:
        b = into(fn, [1, 2, 3])
        assert get_expectedlen(b) == 3
        assert len(b) == 3
        shutil.rmtree(fn)


def test_resource_nd_carray():
    with tmpbcolz(dshape='10 * 10 * 10 * int32') as r:
        assert isinstance(r, carray)
        assert r.dtype == 'i4'
        assert r.shape == (10, 10, 10)


y = np.array([('Alice', 100), ('Bob', 200)],
            dtype=[('name', 'S7'), ('amount', 'i4')])

def test_convert_numpy_to_ctable():
    b = convert(ctable, y)
    assert isinstance(b, ctable)
    assert eq(b[:], y)


def test_resource_existing_ctable():
    with tmpfile('.bcolz') as fn:
        r = into(fn, y)
        r.flush()

        r2 = resource(fn)
        assert eq(r2[:], y)

        shutil.rmtree(fn)


def test_drop():
    with tmpbcolz(dshape='var * {name: string[5, "ascii"], balance: int32}') as b:
        assert os.path.exists(b.rootdir)
        drop(b)
        assert not os.path.exists(b.rootdir)


def test_resource_shape():
    with tmpbcolz(dshape='10 * int') as b:
        assert b.shape == (10,)
    with tmpbcolz(dshape='10 * 10 * int') as b:
        assert b.shape == (10, 10)
    with tmpbcolz(dshape='var * 10 * int') as b:
        assert b.shape == (0, 10)


def test_csv_to_bcolz():
    with filetext('name,runway,takeoff,datetime_nearest_close\n'
                  'S28,28,TRUE,A\n'
                  'S16,16,TRUE,Q\n'
                  'L14,14,FALSE,I', extension='csv') as src:
        with tmpfile('bcolz') as tgt:
            bc = into(tgt, src)
            assert len(bc) == 3



from __future__ import absolute_import, division, print_function

import os
import networkx as nx

from math import log
from toolz import pluck
from .convert import convert, ooc_types
from .append import append
from .utils import cls_name


# Write out as dot
def dot_graph(filename='conversions'):
    # Edges from Convert
    dg = nx.DiGraph()
    for a, b in convert.graph.edges():
        cost = convert.graph.edge[a][b]['cost']
        dg.add_edge(cls_name(a), cls_name(b),
                    cost=cost,
                    penwidth=max(log(1./(cost + 0.06)), 1))


    # Edges from Append
    for a, b in append.funcs:
        if b is not object and a != b:
            dg.add_edge(cls_name(b), cls_name(a), color='blue')


    # Color edges
    for n in convert.graph.nodes() + list(pluck(0, append.funcs)):
        if issubclass(n, tuple(ooc_types)):
            dg.node[cls_name(n)]['color'] = 'red'

    # Convert to pydot
    p = nx.to_pydot(dg)

    p.set_overlap(False)
    p.set_splines(True)

    with open(filename + '.dot', 'w') as f:
        f.write(p.to_string())

    os.system('neato -Tpdf %s.dot -o %s.pdf' % (filename, filename))
    print("Writing graph to %s.pdf" % filename)
    os.system('neato -Tpng %s.dot -o %s.png' % (filename, filename))
    print("Writing graph to %s.png" % filename)

if __name__ == '__main__':
    dot_graph()



from __future__ import absolute_import, division, print_function

import sas7bdat
from sas7bdat import SAS7BDAT

import datashape
from datashape import discover, dshape, var, Record, date_, datetime_
from collections import Iterator
import pandas as pd
from .pandas import coerce_datetimes
from ..append import append
from ..convert import convert
from ..resource import resource


@resource.register('.+\.(sas7bdat)')
def resource_sas(uri, **kwargs):
    return SAS7BDAT(uri, **kwargs)


@discover.register(SAS7BDAT)
def discover_sas(f, **kwargs):
    lines = f.readlines()
    next(lines) # burn header
    ln = next(lines)
    types = map(discover, ln)
    names = [col.name.decode("utf-8") for col in f.header.parent.columns]
    return var * Record(list(zip(names, types)))


@convert.register(pd.DataFrame, SAS7BDAT, cost=4.0)
def sas_to_DataFrame(s, dshape=None, **kwargs):
    df = s.to_data_frame()
    if any(typ in (date_, datetime_) for typ in dshape.measure.types):
        df = coerce_datetimes(df)
    names = [col.decode('utf-8') for col in s.column_names]
    df = df[names]  # Reorder names to match sasfile
    return df


@convert.register(Iterator, SAS7BDAT, cost=1.0)
def sas_to_iterator(s, **kwargs):
    lines = s.readlines()
    if not s.skip_header:
        next(lines)  # burn
    return lines



from __future__ import absolute_import, division, print_function

import pytest
pymongo = pytest.importorskip('pymongo')

import os
from odo import discover, convert, append, resource, dshape, odo
from toolz import pluck
from copy import deepcopy
from bson.objectid import ObjectId

@pytest.fixture(scope='module')
def mongo_host_port():
    import os
    return (os.environ.get('MONGO_IP', 'localhost'),
            os.environ.get('MONGO_PORT', 27017))

@pytest.fixture(scope='module')
def conn(mongo_host_port):
    host, port = mongo_host_port
    try:
        return pymongo.MongoClient(host=host, port=port)
    except pymongo.errors.ConnectionFailure:
        pytest.skip('No mongo server running')


@pytest.yield_fixture
def db(conn):
    try:
        yield conn._test_db
    finally:
        conn.drop_database('_test_db')


@pytest.fixture
def raw_bank():
    return [
        {'name': 'Alice', 'amount': 100},
        {'name': 'Alice', 'amount': 200},
        {'name': 'Bob', 'amount': 100},
        {'name': 'Bob', 'amount': 200},
        {'name': 'Bob', 'amount': 300}
    ]


@pytest.yield_fixture
def bank(db, raw_bank):
    db.bank.insert(deepcopy(raw_bank))
    try:
        yield db.bank
    finally:
        db.drop_collection('bank')


@pytest.yield_fixture
def empty_bank(db):
    try:
        yield db.empty_bank
    finally:
        db.drop_collection('empty_bank')


ds = dshape('var * {name: string, amount: int}')


def test_discover(bank, raw_bank):
    assert discover(bank) == discover(raw_bank)


def test_discover_empty_db(db):
    # NOTE: nothing has been added to the database because our fixtures create
    # a new db and collections for each test that depends on them, so
    # 'system.indexes' won't exist here yet
    # The same is true in test_resource_db
    assert discover(db).measure.names == []


def test_discover_db(bank, db):
    assert 'bank' in set(discover(db).measure.names)


def test_resource_db(mongo_host_port):
    db = resource('mongodb://{}:{}/_test_db'.format(*mongo_host_port))
    assert db.name == '_test_db'
    assert discover(db).measure.names == []


def test_resource_collection(mongo_host_port):
    host, port = mongo_host_port
    coll = resource('mongodb://{}:{}/db::mycoll'.format(*mongo_host_port))
    assert coll.name == 'mycoll'
    assert coll.database.name == 'db'
    assert coll.database.connection.host == host
    assert coll.database.connection.port == port


def test_append_convert(empty_bank, raw_bank):
    ds = discover(raw_bank)
    assert set(ds.measure.names) == {'name', 'amount'}

    append(empty_bank, raw_bank, dshape=ds)
    assert odo(empty_bank, list, dshape=ds) == list(
        pluck(ds.measure.names, raw_bank)
    )


@pytest.yield_fixture
def multiple_object_ids(db):
    data = [
        {'x': 1, 'y': 2, 'other': ObjectId('1' * 24)},
        {'x': 3, 'y': 4, 'other': ObjectId('2' * 24)}
    ]
    db.multiple_object_ids.insert(data)
    try:
        yield db.multiple_object_ids
    finally:
        db.drop_collection('multiple_object_ids')


def test_multiple_object_ids(multiple_object_ids):
    assert discover(multiple_object_ids) == dshape('2 * {x: int64, y: int64}')
    assert convert(list, multiple_object_ids) == [(1, 2), (3, 4)]



from __future__ import absolute_import, division, print_function

from toolz import memoize, merge, partition_all
from multipledispatch import MDNotImplementedError
import re
import os

from .csv import CSV
from .json import JSON, JSONLines
from .text import TextFile
import pandas as pd
import uuid
import datashape
import sqlalchemy as sa
from datashape import discover
from datashape import coretypes as ct
from collections import namedtuple, Iterator
from contextlib import contextmanager
from .ssh import SSH
from .sql import metadata_of_engine
from ..utils import tmpfile, sample, ignoring, raises
from ..temp import Temp
from ..append import append
from ..convert import convert
from ..chunks import chunks
from ..resource import resource
from ..directory import Directory
from ..compatibility import unicode


with ignoring(ImportError):
    from pywebhdfs.webhdfs import PyWebHdfsClient
    from pywebhdfs.errors import FileNotFound


class _HDFS(object):
    """ Parent class for data on Hadoop File System

    Examples
    --------

    >>> HDFS(CSV)('/path/to/file.csv', host='54.91.255.255',
    ...           port=14000, user='hdfs')  # doctest: +SKIP

    Alternatively use resource strings

    >>> resource('hdfs://hdfs@54.91.255.255:/path/to/file.csv')  # doctest: +SKIP
    """
    def __init__(self, *args, **kwargs):
        hdfs = kwargs.get('hdfs', None)
        host = kwargs.get('host', None)
        user = (kwargs.get('user_name') or
                kwargs.get('user') or
                kwargs.get('username') or None)
        port = str(kwargs.get('port', 14000))
        if not hdfs and (host and port and user):
            hdfs = PyWebHdfsClient(host=host, port=str(port),
                                   user_name=user)

        if hdfs is None:
            raise ValueError("No HDFS credentials found.\n"
                    "Either supply a PyWebHdfsClient instance or keywords\n"
                    "   host=, port=, user=")
        self.hdfs = hdfs

        self.subtype.__init__(self, *args, **kwargs)


@memoize
def HDFS(cls):
    return type('HDFS(%s)' % cls.__name__, (_HDFS, cls), {'subtype':  cls})


@sample.register(_HDFS)
@contextmanager
def sample_hdfs_csv(data, length=10000):
    sample = data.hdfs.read_file(data.path.lstrip('/'), length=length)
    with tmpfile(data.canonical_extension) as fn:
        with open(fn, 'w') as f:
            f.write(sample)

        yield fn


@discover.register(HDFS(JSON))
@discover.register(HDFS(JSONLines))
@discover.register(HDFS(TextFile))
def discover_hdfs_file(data, **kwargs):
    with sample(data) as fn:
        result = discover(data.subtype(fn, **kwargs))
    return result


@discover.register(HDFS(CSV))
def discover_hdfs_csv(data, length=10000, **kwargs):
    with sample(data, length=length) as fn:
        result = discover(CSV(fn, encoding=data.encoding,
                                  dialect=data.dialect,
                                  has_header=data.has_header))
    return result


@sample.register(HDFS(Directory(CSV)))
@contextmanager
def sample_hdfs_directory_csv(data, **kwargs):
    files = data.hdfs.list_dir(data.path.lstrip('/'))
    one_file = data.path + '/' + files['FileStatuses']['FileStatus'][0]['pathSuffix']
    csv = HDFS(CSV)(one_file, hdfs=data.hdfs)
    with sample(csv, **kwargs) as fn:
        yield fn


@discover.register(HDFS(Directory(CSV)))
def discover_hdfs_directory(data, length=10000, **kwargs):
    with sample(data, length=length) as fn:
        o = data.container(fn, **data.kwargs)
        result = discover(o)
    return result

"""
Hive Tables
===========

Hive acts a bit differently from other databases that we interact with through
SQL.  As a result we need to special case a lot of code.

Foremost, a database is just a directory on HDFS holding something like a CSV
file (or Avro, Parquet, etc..)  As a result when we construct a Table we
actually have to know a lot about our CSV files (e.g. delimiter.)

This breaks the current odo model a bit because we usually create things with
`resource` and then `append` on to them.  Here we need to know both at the
same time.  Enter a convenient hack, a token for a proxy table
"""

TableProxy = namedtuple('TableProxy', 'engine,name,stored_as')

"""
resource('hive://...::tablename') now gives us one of these.  The
subsequent call to `append` does the actual creation.

We're looking for better solutions.  For the moment, this works.
"""

@resource.register('hive://.+::.+', priority=16)
def resource_hive_table(uri, stored_as='TEXTFILE', external=True, dshape=None, **kwargs):
    if dshape:
        dshape = datashape.dshape(dshape)
    uri, table = uri.split('::')
    engine = resource(uri)
    metadata = metadata_of_engine(engine)

    # If table exists then return it
    with ignoring(sa.exc.NoSuchTableError):
        return sa.Table(table, metadata, autoload=True,
                        autoload_with=engine)

    # Enough information to make an internal table
    if dshape and (not external or external and kwargs.get('path')):
        table_type = 'EXTERNAL' if external else ''
        statement = create_hive_statement(table, dshape,
                db_name=engine.url.database, stored_as=stored_as,
                table_type=table_type, **kwargs)
        with engine.connect() as conn:
            conn.execute(statement)
        return sa.Table(table, metadata, autoload=True,
                        autoload_with=engine)

    else:
        return TableProxy(engine, table, stored_as)


hive_types = {
        ct.int8: 'TINYINT',
        ct.int16: 'SMALLINT',
        ct.int32: 'INT',
        ct.int64: 'BIGINT',
        ct.float32: 'FLOAT',
        ct.float64: 'DOUBLE',
        ct.date_: 'DATE',
        ct.datetime_: 'TIMESTAMP',
        ct.string: 'STRING',
        ct.bool_: 'BOOLEAN'}


def dshape_to_hive(ds):
    """ Convert datashape measure to Hive dtype string

    >>> dshape_to_hive('var * {name: string, balance: int32}')
    ['name  STRING', 'balance  INT']
    >>> dshape_to_hive('int16')
    'SMALLINT'
    >>> dshape_to_hive('?int32')  # Ignore option types
    'INT'
    >>> dshape_to_hive('string[256]')
    'VARCHAR(256)'
    """
    if isinstance(ds, (str, unicode)):
        ds = datashape.dshape(ds)
    if isinstance(ds, ct.DataShape):
        ds = ds.measure
    if isinstance(ds, ct.Record):
        columns = [(name, dshape_to_hive(typ))
                for name, typ in zip(ds.measure.names, ds.measure.types)]
        return ['%s  %s' % col for col in columns]
    if isinstance(ds, ct.Option):
        ds = ds.ty
    if isinstance(ds, ct.String):
        if ds.fixlen:
            return 'VARCHAR(%d)' % ds.fixlen
        else:
            return 'STRING'
    if ds in hive_types:
        return hive_types[ds]
    raise NotImplementedError("No Hive dtype known for %s" % ds)


def create_hive_statement(tbl_name, dshape, path=None, table_type='',
        db_name='default', stored_as='TEXTFILE', **kwargs):
    """ Generic CREATE TABLE statement for hive

    Parameters
    ----------

    tbl_name : string
        Specifies table name "mytable"
    dshape : DataShape
        Datashape of the desired table
    path : string (optional)
        Location of data
    table_type : string (optional)
        Table Modifier like EXTERNAL or LOCAL
    db_name : string
        Specifies database name.  Defaults to "default"
    stored_as: string
        Target storage format like TEXTFILE or PARQUET

    **kwargs: keyword arguments dict
        CSV dialect with keys delimiter, has_header, etc.

    Example
    -------

    >>> from datashape import dshape
    >>> ds = dshape('var * {name: string, balance: int64, when: datetime}')
    >>> print(create_hive_statement('accounts', ds, delimiter=','))  # doctest: +NORMALIZE_WHITESPACE
    CREATE  TABLE default.accounts (
            name  STRING,
         balance  BIGINT,
            when  TIMESTAMP
    )
    ROW FORMAT DELIMITED
        FIELDS TERMINATED BY ','
    STORED AS TEXTFILE

    >>> print(create_hive_statement('accounts', ds, delimiter=',',
    ...         has_header=True, path='/data/accounts/', table_type='EXTERNAL'))  # doctest: +NORMALIZE_WHITESPACE
    CREATE EXTERNAL TABLE default.accounts (
            name  STRING,
         balance  BIGINT,
            when  TIMESTAMP
    )
    ROW FORMAT DELIMITED
        FIELDS TERMINATED BY ','
    STORED AS TEXTFILE
    LOCATION '/data/accounts/'
    TBLPROPERTIES ("skip.header.line.count"="1")

    >>> print(create_hive_statement('accounts', ds, stored_as='PARQUET'))  # doctest: +NORMALIZE_WHITESPACE
    CREATE  TABLE default.accounts (
            name  STRING,
         balance  BIGINT,
            when  TIMESTAMP
    )
    STORED AS PARQUET
    """
    if db_name:
        db_name = db_name + '.'

    if not table_type:
        table_type = ''

    # Column names and types from datashape
    assert isinstance(dshape.measure, ct.Record)
    columns = dshape_to_hive(dshape)
    column_text = ',\n    '.join('%20s  %s' % tuple(col.split())
                                 for col in columns).lstrip()

    statement = """
        CREATE {table_type} TABLE {db_name}{tbl_name} (
            {column_text}
        )
        """

    if 'delimiter' in kwargs:
        statement += """
        ROW FORMAT DELIMITED
            FIELDS TERMINATED BY '{delimiter}'
        """

    statement += """
        STORED AS {stored_as}
        """

    if path:
        statement = statement +"""
        LOCATION '{path}'
        """

    if kwargs.get('has_header'):
        statement = statement + """
        TBLPROPERTIES ("skip.header.line.count"="1")"""

    return statement.format(**merge(kwargs, locals())).strip('\n')


"""
Load Data from HDFS or SSH into Hive
====================================

We push types like HDFS(CSV) and HDFS(Directory(CSV)) into Hive tables.  This
requires that we bring a bit of the CSV file locally, inspect it (sniff for csv
dialect), generate the appropriate CREATE EXTERNAL TABLE command, and then
execute.
"""


@append.register(TableProxy, HDFS(Directory(CSV)))
def create_new_hive_table_from_csv(tbl, data, dshape=None, path=None, **kwargs):
    """
    Create new Hive table from directory of CSV files on HDFS

    Actually executes command.
    """
    table_type = 'EXTERNAL'

    if not dshape:
        dshape = discover(data)

    if tbl.engine.dialect.name == 'hive':
        statement = create_hive_statement(tbl.name, dshape,
                                path=data.path,
                                db_name = str(tbl.engine.url).split('/')[-1],
                                table_type=table_type,
                                **dialect_of(data))
    else:
        raise NotImplementedError("Don't know how to migrate directory of csvs"
                " on HDFS to database of dialect %s" % tbl.engine.dialect.name)

    with tbl.engine.connect() as conn:
        conn.execute(statement)

    metadata = metadata_of_engine(tbl.engine)
    tbl2 = sa.Table(tbl.name, metadata, autoload=True,
            autoload_with=tbl.engine)
    return tbl2


@append.register(TableProxy, (HDFS(CSV), Temp(HDFS(CSV))))
def create_new_hive_table_from_csv_file(tbl, data, dshape=None, path=None, **kwargs):
    raise ValueError(
        "Can not create a new Hive table from a single CSV file on HDFS.\n"
        "Instead try loading a complete directory or base your data outside of"
        " HDFS")


@append.register(TableProxy, (SSH(CSV), SSH(Directory(CSV))))
def append_remote_csv_to_new_table(tbl, data, dshape=None, **kwargs):
    if not dshape:
        dshape = discover(data)

    if tbl.engine.dialect.name == 'hive':
        statement = create_hive_statement(tbl.name, dshape,
                                db_name = str(tbl.engine.url).split('/')[-1],
                                **dialect_of(data))
    else:
        raise NotImplementedError("Don't know how to migrate directory of csvs"
                " on Local disk to database of dialect %s" % tbl.engine.dialect.name)

    with tbl.engine.connect() as conn:
        conn.execute(statement)

    metadata = metadata_of_engine(tbl.engine)
    tbl2 = sa.Table(tbl.name, metadata, autoload=True,
            autoload_with=tbl.engine)

    return append(tbl2, data, **kwargs)


@append.register(TableProxy, object)
def append_anything_to_tableproxy(tbl, data, **kwargs):
    return append(tbl, convert(Temp(SSH(CSV)), data, **kwargs), **kwargs)


@append.register(sa.Table, (SSH(CSV), SSH(Directory(CSV))))
def append_remote_csv_to_table(tbl, csv, **kwargs):
    """
    Load Remote data into existing Hive table
    """
    path = csv.path
    if path[0] != '/':
        path = '/home/%s/%s' % (csv.auth['username'], csv.path)

    if tbl.bind.dialect.name == 'hive':
        statement =('LOAD DATA LOCAL INPATH "%(path)s" INTO TABLE %(tablename)s'
                     % {'path': path, 'tablename': tbl.name})
    else:
        raise NotImplementedError("Don't know how to migrate csvs on remote "
                  "disk to database of dialect %s" % tbl.engine.dialect.name)
    with tbl.bind.connect() as conn:
        conn.execute(statement)
    return tbl


@append.register(sa.Table, (HDFS(CSV), HDFS(Directory(CSV)), Temp(HDFS(CSV))))
def append_hdfs_csv_to_table(tbl, csv, **kwargs):
    """
    Load Remote data into existing Hive table
    """
    if tbl.bind.dialect.name != 'hive':
        raise NotImplementedError("Don't know how to migrate csvs on remote "
                  "disk to database of dialect %s" % tbl.engine.dialect.name)

    statement =('LOAD DATA INPATH "%(path)s" INTO TABLE %(tablename)s'
                 % {'path': csv.path, 'tablename': tbl.name})
    with tbl.bind.connect() as conn:
        conn.execute(statement)
    return tbl

@append.register(TextFile, HDFS(TextFile))
@append.register(JSONLines, HDFS(JSONLines))
@append.register(JSON, HDFS(JSON))
@append.register(CSV, HDFS(CSV))
def append_hdfs_file_to_local(target, source, **kwargs):
    text = source.hdfs.read_file(source.path.lstrip('/'), **kwargs)
    with open(target.path, 'w') as f:
        f.write(text)
    return target


@convert.register(Temp(TextFile), HDFS(TextFile))
@convert.register(Temp(JSONLines), HDFS(JSONLines))
@convert.register(Temp(JSON), HDFS(JSON))
@convert.register(Temp(CSV), HDFS(CSV))
def convert_hdfs_file_to_temp_local(source, **kwargs):
    ext = os.path.splitext(source.path)[1].strip('.')
    fn = '.%s.%s' % (str(uuid.uuid1()), ext)
    tmp = Temp(source.subtype)(fn)
    return append(tmp, source, **kwargs)


@append.register(HDFS(TextFile), TextFile)
@append.register(HDFS(JSONLines), JSONLines)
@append.register(HDFS(JSON), JSON)
@append.register(HDFS(CSV), (CSV, Temp(CSV)))
def append_local_file_to_hdfs(target, source, blocksize=100000, **kwargs):
    if raises(FileNotFound,
              lambda: target.hdfs.list_dir(target.path.lstrip('/'))):
        target.hdfs.create_file(target.path.lstrip('/'), '')

    with open(source.path, 'r') as f:
        blocks = partition_all(blocksize, f)
        for block in blocks:
            target.hdfs.append_file(target.path.lstrip('/'), ''.join(block))

    return target


import csv
sniffer = csv.Sniffer()


def dialect_of(data, **kwargs):
    """ CSV dialect of a CSV file stored in SSH, HDFS, or a Directory. """
    keys = set(['delimiter', 'doublequote', 'escapechar', 'lineterminator',
                'quotechar', 'quoting', 'skipinitialspace', 'strict', 'has_header'])
    if isinstance(data, (HDFS(CSV), SSH(CSV))):
        with sample(data) as fn:
            d = dialect_of(CSV(fn, **data.dialect))
    elif isinstance(data, (HDFS(Directory(CSV)), SSH(Directory(CSV)))):
        with sample(data) as fn:
            d = dialect_of(CSV(fn, **data.kwargs))
    elif isinstance(data, Directory(CSV)):
        d = dialect_of(next(data))
    else:
        assert isinstance(data, CSV)

        # Get sample text
        with open(data.path, 'r') as f:
            text = f.read()

        result = dict()

        d = sniffer.sniff(text)
        d = dict((k, getattr(d, k)) for k in keys if hasattr(d, k))

        if data.has_header is None:
            d['has_header'] = sniffer.has_header(text)
        else:
            d['has_header'] = data.has_header

        d.update(data.dialect)

    d.update(kwargs)
    d = dict((k, v) for k, v in d.items() if k in keys)

    return d



types_by_extension = {'csv': CSV, 'json': JSONLines, 'txt': TextFile,
                      'log': TextFile}

hdfs_pattern = '(((?P<user>[a-zA-Z]\w*)@)?(?P<host>[\w.-]*)?(:(?P<port>\d+))?:)?(?P<path>[/\w.*-]+)'

@resource.register('hdfs://.*', priority=16)
def resource_hdfs(uri, **kwargs):
    if 'hdfs://' in uri:
        uri = uri[len('hdfs://'):]

    d = re.match(hdfs_pattern, uri).groupdict()
    d = dict((k, v) for k, v in d.items() if v is not None)
    path = d.pop('path')

    kwargs.update(d)

    try:
        subtype = types_by_extension[path.split('.')[-1]]
        if '*' in path:
            subtype = Directory(subtype)
            path = path.rsplit('/', 1)[0] + '/'
    except KeyError:
        subtype = type(resource(path))

    return HDFS(subtype)(path, **kwargs)


@append.register(HDFS(TextFile), (Iterator, object))
@append.register(HDFS(JSON), (list, object))
@append.register(HDFS(CSV), (chunks(pd.DataFrame), pd.DataFrame, object))
def append_object_to_hdfs(target, source, **kwargs):
    tmp = convert(Temp(target.subtype), source, **kwargs)
    return append(target, tmp, **kwargs)


@append.register(HDFS(TextFile), SSH(TextFile))
@append.register(HDFS(JSONLines), SSH(JSONLines))
@append.register(HDFS(JSON), SSH(JSON))
@append.register(HDFS(CSV), SSH(CSV))
def append_remote_file_to_hdfs(target, source, **kwargs):
    raise MDNotImplementedError()


@append.register(HDFS(TextFile), HDFS(TextFile))
@append.register(HDFS(JSONLines), HDFS(JSONLines))
@append.register(HDFS(JSON), HDFS(JSON))
@append.register(HDFS(CSV), HDFS(CSV))
def append_hdfs_file_to_hdfs_file(target, source, **kwargs):
    raise MDNotImplementedError()


@append.register(SSH(TextFile), HDFS(TextFile))
@append.register(SSH(JSONLines), HDFS(JSONLines))
@append.register(SSH(JSON), HDFS(JSON))
@append.register(SSH(CSV), HDFS(CSV))
def append_hdfs_file_to_remote(target, source, **kwargs):
    raise MDNotImplementedError()



from __future__ import absolute_import, division, print_function

from datashape import discover
from datashape.dispatch import dispatch
from ..append import append
from ..convert import convert, ooc_types
from ..resource import resource
from ..chunks import chunks
from ..utils import tmpfile

import os

import numpy as np
import tables

from toolz import first

import datashape

import shutil

__all__ = ['PyTables']


@discover.register((tables.Array, tables.Table))
def discover_tables_node(n):
    return datashape.from_numpy(n.shape, n.dtype)


@discover.register(tables.Node)
def discover_tables_node(n):
    return discover(n._v_children)  # subclasses dict


@discover.register(tables.File)
def discover_tables_node(f):
    return discover(f.getNode('/'))


@append.register((tables.Array, tables.Table), np.ndarray)
def numpy_to_pytables(t, x, **kwargs):
    t.append(x)
    return x


@append.register((tables.Array, tables.Table), object)
def append_h5py(dset, x, **kwargs):
    return append(dset, convert(chunks(np.ndarray), x, **kwargs), **kwargs)


@convert.register(np.ndarray, tables.Table, cost=3.0)
def pytables_to_numpy(t, **kwargs):
    return t[:]


@convert.register(chunks(np.ndarray), tables.Table, cost=3.0)
def pytables_to_numpy_chunks(t, chunksize=2**20, **kwargs):
    def load():
        for i in range(0, t.shape[0], chunksize):
            yield t[i: i + chunksize]
    return chunks(np.ndarray)(load)


def dtype_to_pytables(dtype):
    """ Convert NumPy dtype to PyTable descriptor

    Examples
    --------
    >>> from tables import Int32Col, StringCol, Time64Col
    >>> dt = np.dtype([('name', 'S7'), ('amount', 'i4'), ('time', 'M8[us]')])
    >>> dtype_to_pytables(dt)  # doctest: +SKIP
    {'amount': Int32Col(shape=(), dflt=0, pos=1),
     'name': StringCol(itemsize=7, shape=(), dflt='', pos=0),
     'time': Time64Col(shape=(), dflt=0.0, pos=2)}
    """
    d = {}
    for pos, name in enumerate(dtype.names):
        dt, _ = dtype.fields[name]
        if issubclass(dt.type, np.datetime64):
            tdtype = tables.Description({name: tables.Time64Col(pos=pos)}),
        else:
            tdtype = tables.descr_from_dtype(np.dtype([(name, dt)]))
        el = first(tdtype)
        getattr(el, name)._v_pos = pos
        d.update(el._v_colobjects)
    return d


def PyTables(path, datapath, dshape=None, **kwargs):
    """Create or open a ``tables.Table`` object.

    Parameters
    ----------
    path : str
        Path to a PyTables HDF5 file.
    datapath : str
        The name of the node in the ``tables.File``.
    dshape : str or datashape.DataShape
        DataShape to use to create the ``Table``.

    Returns
    -------
    t : tables.Table

    Examples
    --------
    >>> from odo.utils import tmpfile
    >>> # create from scratch
    >>> with tmpfile('.h5') as f:
    ...     t = PyTables(filename, '/bar',
    ...                  dshape='var * {volume: float64, planet: string[10, "A"]}')
    ...     data = [(100.3, 'mars'), (100.42, 'jupyter')]
    ...     t.append(data)
    ...     t[:]  # doctest: +SKIP
    ...
    array([(100.3, b'mars'), (100.42, b'jupyter')],
          dtype=[('volume', '<f8'), ('planet', 'S10')])
    """
    def possibly_create_table(filename, dtype):
        f = tables.open_file(filename, mode='a')
        try:
            if datapath not in f:
                if dtype is None:
                    raise ValueError('dshape cannot be None and datapath not'
                                     ' in file')
                else:
                    f.create_table('/', datapath.lstrip('/'), description=dtype)
        finally:
            f.close()

    if dshape:
        if isinstance(dshape, str):
            dshape = datashape.dshape(dshape)
        if dshape[0] == datashape.var:
            dshape = dshape.subshape[0]
        dtype = dtype_to_pytables(datashape.to_numpy_dtype(dshape))
    else:
        dtype = None

    if os.path.exists(path):
        possibly_create_table(path, dtype)
    else:
        with tmpfile('.h5') as filename:
            possibly_create_table(filename, dtype)
            shutil.copyfile(filename, path)
    return tables.open_file(path, mode='a').get_node(datapath)


@resource.register('pytables://.+', priority=11)
def resource_pytables(path, datapath, **kwargs):
    return PyTables(path[len('pytables://'):], datapath, **kwargs)


@dispatch((tables.Table, tables.Array))
def drop(t):
    t.remove()


@dispatch(tables.File)
def drop(f):
    f.close()
    os.remove(f.filename)


ooc_types |= set((tables.Table, tables.Array))



print("import: 'odo'")
import odo

print("import: 'odo.backends'")
import odo.backends

print("import: 'odo.backends.tests'")
import odo.backends.tests

print("import: 'odo.tests'")
import odo.tests




from __future__ import absolute_import, division, print_function

import warnings
import functools
import inspect
import datetime
import tempfile
import os
import shutil
import numpy as np

from contextlib import contextmanager
from multiprocessing.pool import ThreadPool

from multipledispatch import Dispatcher

from datashape import dshape, Record
from datashape.discovery import is_zero_time

from toolz import pluck, get, curry, keyfilter

from .compatibility import unicode

sample = Dispatcher('sample')


def iter_except(func, exception, first=None):
    """Call a `func` repeatedly until `exception` is raised. Optionally call
    `first` first.

    Parameters
    ----------
    func : callable
        Repeatedly call this until `exception` is raised.
    exception : Exception
        Stop calling `func` when this is raised.
    first : callable, optional, default ``None``
        Call this first if it isn't ``None``.

    Examples
    --------
    >>> x = {'a': 1, 'b': 2}
    >>> def iterate():
    ...     yield 'a'
    ...     yield 'b'
    ...     yield 'c'
    ...
    >>> keys = iterate()
    >>> diter = iter_except(lambda: x[next(keys)], KeyError)
    >>> list(diter)
    [1, 2]

    Notes
    -----
    * Taken from https://docs.python.org/2/library/itertools.html#recipes
    """
    try:
        if first is not None:
            yield first()
        while 1:  # True isn't a reserved word in Python 2.x
            yield func()
    except exception:
        pass


def ext(filename):
    _, e = os.path.splitext(filename)
    return e.lstrip(os.extsep)


def raises(err, lamda):
    try:
        lamda()
        return False
    except err:
        return True


def expand_tuples(L):
    """

    >>> expand_tuples([1, (2, 3)])
    [(1, 2), (1, 3)]

    >>> expand_tuples([1, 2])
    [(1, 2)]
    """
    if not L:
        return [()]
    elif not isinstance(L[0], tuple):
        rest = expand_tuples(L[1:])
        return [(L[0],) + t for t in rest]
    else:
        rest = expand_tuples(L[1:])
        return [(item,) + t for t in rest for item in L[0]]


@contextmanager
def tmpfile(extension='', dir=None):
    extension = '.' + extension.lstrip('.')
    handle, filename = tempfile.mkstemp(extension, dir=dir)
    os.close(handle)
    os.remove(filename)

    try:
        yield filename
    finally:
        if os.path.exists(filename):
            if os.path.isdir(filename):
                shutil.rmtree(filename)
            else:
                try:
                    os.remove(filename)
                except OSError:
                    # sometimes we can't remove a generated temp file
                    pass


def keywords(func):
    """ Get the argument names of a function

    >>> def f(x, y=2):
    ...     pass

    >>> keywords(f)
    ['x', 'y']
    """
    if isinstance(func, type):
        return keywords(func.__init__)
    return inspect.getargspec(func).args


def cls_name(cls):
    if 'builtin' in cls.__module__:
        return cls.__name__
    else:
        return cls.__module__.split('.')[0] + '.' + cls.__name__


@contextmanager
def filetext(text, extension='', open=open, mode='w'):
    with tmpfile(extension=extension) as filename:
        f = open(filename, mode=mode)
        try:
            f.write(text)
        finally:
            try:
                f.close()
            except AttributeError:
                pass

        try:
            yield filename
        finally:
            if os.path.exists(filename):
                try:
                    os.remove(filename)
                except OSError:
                    pass


@contextmanager
def filetexts(d, open=open):
    """ Dumps a number of textfiles to disk

    d - dict
        a mapping from filename to text like {'a.csv': '1,1\n2,2'}
    """
    for filename, text in d.items():
        f = open(filename, 'wt')
        try:
            f.write(text)
        finally:
            try:
                f.close()
            except AttributeError:
                pass

    try:
        yield list(d)
    finally:
        for filename in d:
            if os.path.exists(filename):
                try:
                    os.remove(filename)
                except OSError:
                    pass


def normalize_to_date(dt):
    if isinstance(dt, datetime.datetime) and is_zero_time(dt.time()):
        return dt.date()
    else:
        return dt


def assert_allclose(lhs, rhs):
    for tb in map(zip, lhs, rhs):
        for left, right in tb:
            if isinstance(left, (np.floating, float)):
                # account for nans
                assert np.all(np.isclose(left, right, equal_nan=True))
                continue
            if isinstance(left, datetime.datetime):
                left = normalize_to_date(left)
            if isinstance(right, datetime.datetime):
                right = normalize_to_date(right)
            assert left == right


def records_to_tuples(ds, data):
    """ Transform records into tuples

    Examples
    --------
    >>> seq = [{'a': 1, 'b': 10}, {'a': 2, 'b': 20}]
    >>> list(records_to_tuples('var * {a: int, b: int}', seq))
    [(1, 10), (2, 20)]

    >>> records_to_tuples('{a: int, b: int}', seq[0])  # single elements
    (1, 10)

    >>> records_to_tuples('var * int', [1, 2, 3])  # pass through on non-records
    [1, 2, 3]

    See Also
    --------

    tuples_to_records
    """
    if isinstance(ds, (str, unicode)):
        ds = dshape(ds)
    if isinstance(ds.measure, Record) and len(ds.shape) == 1:
        return pluck(ds.measure.names, data, default=None)
    if isinstance(ds.measure, Record) and len(ds.shape) == 0:
        return get(ds.measure.names, data)
    if not isinstance(ds.measure, Record):
        return data
    raise NotImplementedError()


def tuples_to_records(ds, data):
    """ Transform tuples into records

    Examples
    --------
    >>> seq = [(1, 10), (2, 20)]
    >>> list(tuples_to_records('var * {a: int, b: int}', seq))  # doctest: +SKIP
    [{'a': 1, 'b': 10}, {'a': 2, 'b': 20}]

    >>> tuples_to_records('{a: int, b: int}', seq[0])  # doctest: +SKIP
    {'a': 1, 'b': 10}

    >>> tuples_to_records('var * int', [1, 2, 3])  # pass through on non-records
    [1, 2, 3]

    See Also
    --------
    records_to_tuples
    """
    if isinstance(ds, (str, unicode)):
        ds = dshape(ds)
    if isinstance(ds.measure, Record) and len(ds.shape) == 1:
        names = ds.measure.names
        return (dict(zip(names, tup)) for tup in data)
    if isinstance(ds.measure, Record) and len(ds.shape) == 0:
        names = ds.measure.names
        return dict(zip(names, data))
    if not isinstance(ds.measure, Record):
        return data
    raise NotImplementedError()


@contextmanager
def ignoring(*exceptions):
    try:
        yield
    except exceptions:
        pass


def into_path(*path):
    """ Path to file in into directory

    >>> into_path('backends', 'tests', 'myfile.csv')  # doctest: +SKIP
    '/home/user/odo/odo/backends/tests/myfile.csv'
    """
    import odo
    return os.path.join(os.path.dirname(odo.__file__), *path)


@curry
def pmap(f, iterable):
    """Map `f` over `iterable` in parallel using a ``ThreadPool``.
    """
    p = ThreadPool()
    try:
        result = p.map(f, iterable)
    finally:
        p.terminate()
    return result


@curry
def write(triple, writer):
    """Write a file using the input from `gentemp` using `writer` and return
    its index and filename.

    Parameters
    ----------
    triple : tuple of int, str, str
        The first element is the index in the set of chunks of a file, the
        second element is the path to write to, the third element is the data
        to write.

    Returns
    -------
    i, filename : int, str
        File's index and filename. This is used to return the index and
        filename after splitting files.

    Notes
    -----
    This could be adapted to write to an already open handle, which would
    allow, e.g., multipart gzip uploads. Currently we open write a new file
    every time.
    """
    i, filename, data = triple
    with writer(filename, mode='wb') as f:
        f.write(data)
    return i, filename


def gentemp(it, suffix=None, start=0):
    """Yield an index, a temp file, and data for each element in `it`.

    Parameters
    ----------
    it : Iterable
    suffix : str or ``None``, optional
        Suffix to add to each temporary file's name
    start : int, optional
        A integer indicating where to start the numbering of chunks in `it`.
    """
    for i, data in enumerate(it, start=start):  # aws needs parts to start at 1
        with tmpfile('.into') as fn:
            yield i, fn, data


@curry
def split(filename, nbytes, suffix=None, writer=open, start=0):
    """Split a file into chunks of size `nbytes` with each filename containing
    a suffix specified by `suffix`. The file will be written with the ``write``
    method of an instance of `writer`.

    Parameters
    ----------
    filename : str
        The file to split
    nbytes : int
        Split `filename` into chunks of this size
    suffix : str, optional
    writer : callable, optional
        Callable object to use to write the chunks of `filename`
    """
    with open(filename, mode='rb') as f:
        byte_chunks = iter(curry(f.read, nbytes), '')
        return pmap(write(writer=writer),
                    gentemp(byte_chunks, suffix=suffix, start=start))


def filter_kwargs(f, kwargs):
    """Return a dict of valid kwargs for `f` from a subset of `kwargs`

    Examples
    --------
    >>> def f(a, b=1, c=2):
    ...     return a + b + c
    ...
    >>> raw_kwargs = dict(a=1, b=3, d=4)
    >>> f(**raw_kwargs)
    Traceback (most recent call last):
        ...
    TypeError: f() got an unexpected keyword argument 'd'
    >>> kwargs = filter_kwargs(f, raw_kwargs)
    >>> f(**kwargs)
    6
    """
    return keyfilter(keywords(f).__contains__, kwargs)


@curry
def copydoc(from_, to):
    """Copies the docstring from one function to another.

    Parameters
    ----------
    from_ : any
        The object to copy the docstring from.
    to : any
        The object to copy the docstring to.

    Returns
    -------
    to : any
        ``to`` with the docstring from ``from_``
    """
    to.__doc__ = from_.__doc__
    return to

def deprecated(replacement=None):
    """A decorator which can be used to mark functions as deprecated.
    replacement is a callable that will be called with the same args
    as the decorated function.
    """
    def outer(fun):
        msg = "{} is deprecated".format(fun.__name__)
        if replacement is not None:
            msg += "; use {} instead".format(replacement)
        if fun.__doc__ is None:
            fun.__doc__ = msg

        @functools.wraps(fun)
        def inner(*args, **kwargs):
            warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
            return fun(*args, **kwargs)

        return inner
    return outer


def literal_compile(s):
    """Compile a sql expression with bind params inlined as literals.

    Parameters
    ----------
    s : Selectable
        The expression to compile.

    Returns
    -------
    cs : str
        An equivalent sql string.
    """
    return str(s.compile(compile_kwargs={'literal_binds': True}))



from __future__ import absolute_import, division, print_function

from datashape.dispatch import dispatch
from .resource import resource
from .compatibility import unicode


@dispatch(object)
def drop(rsrc):
    """Remove a resource.

    Parameters
    ----------
    rsrc : CSV, SQL, tables.Table, pymongo.Collection
        A resource that will be removed. For example, calling ``drop(csv)`` will
        delete the CSV file.

    Examples
    --------
    >>> # Using SQLite
    >>> from odo import resource, odo
    >>> # create a table called 'tb', in memory
    >>> from datashape import dshape
    >>> ds = dshape('var * {name: string, amount: int}')
    >>> sql = resource('sqlite:///:memory:::tb', dshape=ds)
    >>> odo([('Alice', 100), ('Bob', 200)], sql)
    >>> odo(sql, list)
    [('Alice', 100), ('Bob', 200)]
    >>> drop(sql)
    """
    raise NotImplementedError("drop not implemented for type %r" %
                              type(rsrc).__name__)


@dispatch((str, unicode))
def drop(uri, **kwargs):
    data = resource(uri, **kwargs)
    drop(data)



from __future__ import absolute_import, division, print_function

import pytest
h5py = pytest.importorskip('h5py')

import os
import sys

from distutils.version import LooseVersion

import h5py
from contextlib import contextmanager

from odo.backends.h5py import append, create, resource, discover, convert
from odo.backends.h5py import unicode_dtype

from odo.utils import tmpfile, ignoring
from odo.chunks import chunks
from odo import into, append, convert, discover, drop, odo
import datashape
import h5py
import numpy as np


@contextmanager
def file(x):
    with tmpfile('.hdf5') as fn:
        f = h5py.File(fn)
        data = f.create_dataset('/data', data=x, chunks=True,
                                maxshape=(None,) + x.shape[1:])

        try:
            yield fn, f, data
        finally:
            with ignoring(Exception):
                f.close()


x = np.ones((2, 3), dtype='i4')


def test_drop_group():
    with tmpfile('.hdf5') as fn:
        f = h5py.File(fn)
        try:
            f.create_dataset('/group/data', data=x, chunks=True,
                             maxshape=(None,) + x.shape[1:])
            drop(f['/group'])
            assert '/group' not in f.keys()
        finally:
            with ignoring(Exception):
                f.close()


def test_drop_dataset():
    with tmpfile('.hdf5') as fn:
        f = h5py.File(fn)
        try:
            data = f.create_dataset('/data', data=x, chunks=True,
                                    maxshape=(None,) + x.shape[1:])

            drop(data)
            assert '/data' not in f.keys()
        finally:
            with ignoring(Exception):
                f.close()


def test_drop_file():
    with file(x) as (fn, f, data):
        drop(f)
    assert not os.path.exists(fn)


def test_discover():
    with file(x) as (fn, f, dset):
        assert str(discover(dset)) == str(discover(x))
        assert str(discover(f)) == str(discover({'data': x}))


two_point_five_and_windows_py3 = \
    pytest.mark.skipif(sys.platform == 'win32' and
                       h5py.__version__ == LooseVersion('2.5.0') and
                       sys.version_info[0] == 3,
                       reason=('h5py 2.5.0 issue with varlen string types: '
                               'https://github.com/h5py/h5py/issues/593'))


@two_point_five_and_windows_py3
def test_discover_on_data_with_object_in_record_name():
    data = np.array([(u'a', 1), (u'b', 2)], dtype=[('lrg_object',
                                                    unicode_dtype),
                                                   ('an_int', 'int64')])
    with tmpfile('.hdf5') as fn:
        f = h5py.File(fn)
        try:
            f.create_dataset('/data', data=data)
        except:
            raise
        else:
            assert (discover(f['data']) ==
                    datashape.dshape('2 * {lrg_object: string, an_int: int64}'))
        finally:
            with ignoring(Exception):
                f.close()


def eq(a, b):
    c = a == b
    if isinstance(c, np.ndarray):
        c = c.all()
    return c


def test_append():
    with file(x) as (fn, f, dset):
        append(dset, x)
        assert eq(dset[:], np.concatenate([x, x]))


def test_append_with_uri():
    with file(x) as (fn, f, dset):
        result = odo(dset, '%s::%s' % (fn, dset.name))
        assert eq(result[:], np.concatenate([x, x]))


def test_into_resource():
    with tmpfile('.hdf5') as fn:
        d = into(fn + '::/x', x)
        try:
            assert d.shape == x.shape
            assert eq(d[:], x[:])
        finally:
            d.file.close()


def test_numpy():
    with file(x) as (fn, f, dset):
        assert eq(convert(np.ndarray, dset), x)


def test_chunks():
    with file(x) as (fn, f, dset):
        c = convert(chunks(np.ndarray), dset)
        assert eq(convert(np.ndarray, c), x)


def test_append_chunks():
    with file(x) as (fn, f, dset):
        append(dset, chunks(np.ndarray)([x, x]))

        assert len(dset) == len(x) * 3


def test_create():
    with tmpfile('.hdf5') as fn:
        ds = datashape.dshape('{x: int32, y: {z: 3 * int32}}')
        f = create(h5py.File, dshape='{x: int32, y: {z: 3 * int32}}', path=fn)
        try:
            assert isinstance(f, h5py.File)
            assert f.filename == fn
            assert discover(f) == ds
        finally:
            f.close()


def test_create_partially_present_dataset():
    with tmpfile('.hdf5') as fn:
        ds1 = datashape.dshape('{x: int32}')
        f = create(h5py.File, dshape=ds1, path=fn)

        ds2 = datashape.dshape('{x: int32, y: 5 * int32}')
        f2 = create(h5py.File, dshape=ds2, path=fn)

        try:
            assert f.filename == f2.filename
            assert list(f.keys()) == list(f2.keys())
            assert f['y'].dtype == 'i4'
        finally:
            f.close()
            f2.close()



def test_resource():
    with tmpfile('.hdf5') as fn:
        ds = datashape.dshape('{x: int32, y: 3 * int32}')
        r = resource(fn, dshape=ds)
        r2 = resource(fn + '::/x')

        try:
            assert isinstance(r, h5py.File)
            assert discover(r) == ds
            assert isinstance(r2, h5py.Dataset)
        finally:
            r.close()
            r2.file.close()


def test_resource_with_datapath():
    with tmpfile('.hdf5') as fn:
        ds = datashape.dshape('3 * 4 * int32')
        r = resource(fn + '::/data', dshape=ds)

        try:
            assert isinstance(r, h5py.Dataset)
            assert discover(r) == ds
            assert r.file.filename == fn
            assert r.file['/data'] == r
        finally:
            r.file.close()


def test_resource_with_variable_length():
    with tmpfile('.hdf5') as fn:
        ds = datashape.dshape('var * 4 * int32')
        r = resource(fn + '::/data', dshape=ds)
        try:
            assert r.shape == (0, 4)
        finally:
            r.file.close()


@two_point_five_and_windows_py3
def test_resource_with_option_types():
    with tmpfile('.hdf5') as fn:
        ds = datashape.dshape('4 * {name: ?string, amount: ?int32}')
        r = resource(fn + '::/data', dshape=ds)
        try:
            assert r.shape == (4,)
            assert r.dtype == [('name', 'O'), ('amount', 'f4')]
        finally:
            r.file.close()


def test_resource_with_h5py_protocol():
    with tmpfile('.hdf5') as fn:
        assert isinstance(resource('h5py://' + fn), h5py.File)
    with tmpfile('.xyz') as fn:
        assert isinstance(resource('h5py://' + fn), h5py.File)


def test_copy_with_into():
    with tmpfile('.hdf5') as fn:
        dset = into(fn + '::/data', [1, 2, 3])
        try:
            assert dset.shape == (3,)
            assert eq(dset[:], [1, 2, 3])
        finally:
            dset.file.close()


@two_point_five_and_windows_py3
def test_varlen_dtypes():
    y = np.array([('Alice', 100), ('Bob', 200)],
                dtype=[('name', 'O'), ('amount', 'i4')])
    with tmpfile('.hdf5') as fn:
        dset = into(fn + '::/data', y)
        try:
            assert into(list, dset) == into(list, dset)
        finally:
            dset.file.close()


def test_resource_shape():
    with tmpfile('.hdf5') as fn:
        r = resource(fn+'::/data', dshape='10 * int')
        assert r.shape == (10,)
        r.file.close()
    with tmpfile('.hdf5') as fn:
        r = resource(fn+'::/data', dshape='10 * 10 * int')
        assert r.shape == (10, 10)
        r.file.close()
    with tmpfile('.hdf5') as fn:
        r = resource(fn+'::/data', dshape='var * 10 * int')
        assert r.shape == (0, 10)
        r.file.close()


@pytest.mark.parametrize('ext', ['h5', 'hdf5'])
def test_resource_with_hdfstore_extension_works(ext):
    fn = 'hdfstore:foo.%s' % ext
    with pytest.raises(NotImplementedError):
        odo(fn, np.recarray)
    assert not os.path.exists(fn)



from __future__ import absolute_import, division, print_function

import pytest
bokeh = pytest.importorskip('bokeh')

from odo.backends.bokeh import convert, pd, ColumnDataSource
import pandas.util.testing as tm


df = pd.DataFrame([[100, 'Alice'],
                   [200, 'Bob'],
                   [300, 'Charlie']],
                  columns=['balance', 'name'])


def test_convert_dataframe_to_cds():
    cds = convert(ColumnDataSource, df)
    assert cds.data == {'name': ['Alice', 'Bob', 'Charlie'],
                        'balance': [100, 200, 300]}

    df2 = convert(pd.DataFrame, cds)
    assert isinstance(df2, pd.DataFrame)

    tm.assert_frame_equal(df, df2)





from __future__ import absolute_import, division, print_function

import json
from toolz.curried import map, take, pipe, pluck, get, concat, filter
from collections import Iterator, Iterable
import os
from contextlib import contextmanager

from datashape import discover, var, dshape, Record, DataShape
from datashape import coretypes as ct
from datashape.dispatch import dispatch
import gzip
import datetime
import uuid
from ..append import append
from ..convert import convert, ooc_types
from ..resource import resource
from ..chunks import chunks
from ..temp import Temp
from ..drop import drop
from ..utils import tuples_to_records


class JSON(object):
    """ Proxy for a JSON file

    Parameters
    ----------

    path : str
        Path to file on disk

    See Also
    --------

    JSONLines - Line-delimited JSON
    """
    canonical_extension = 'json'

    def __init__(self, path, **kwargs):
        self.path = path


class JSONLines(object):
    """ Proxy for a line-delimited JSON file

    Each line in the file is a valid JSON entry

    Parameters
    ----------

    path : str
        Path to file on disk

    See Also
    --------

    JSON - Not-line-delimited JSON
    """
    canonical_extension = 'json'

    def __init__(self, path, **kwargs):
        self.path = path


def date_to_datetime_dshape(ds):
    shape = ds.shape
    if isinstance(ds.measure, Record):
        measure = Record([[name, ct.datetime_ if typ == ct.date_ else typ]
            for name, typ in ds.measure.parameters[0]])
    else:
        measure = ds.measure
    return DataShape(*(shape + (measure,)))


@discover.register(JSON)
def discover_json(j, **kwargs):
    data = json_load(j.path)
    ds = discover(data)
    return date_to_datetime_dshape(ds)


def nonempty(line):
    return len(line.strip()) > 0


@discover.register(JSONLines)
def discover_jsonlines(j, n=10, encoding='utf-8', **kwargs):
    with json_lines(j.path, encoding=encoding) as lines:
        data = pipe(lines, filter(nonempty), map(json.loads), take(n), list)

    if len(data) < n:
        ds = discover(data)
    else:
        ds = var * discover(data).subshape[0]
    return date_to_datetime_dshape(ds)



@convert.register(list, (JSON, Temp(JSON)))
def json_to_list(j, dshape=None, **kwargs):
    return json_load(j.path, **kwargs)


@convert.register(Iterator, (JSONLines, Temp(JSONLines)))
def json_lines_to_iterator(j, encoding='utf-8', **kwargs):
    with json_lines(j.path, encoding=encoding) as lines:
        for item in pipe(lines, filter(nonempty), map(json.loads)):
            yield item


@contextmanager
def json_lines(path, encoding='utf-8'):
    """ Return lines of a json-lines file

    Handles compression like gzip """
    if path.split(os.path.extsep)[-1] == 'gz':
        f = gzip.open(path)
        lines = (line.decode(encoding) for line in f)
    else:
        f = open(path)
        lines = f

    try:
        yield lines
    finally:
        f.close()


def json_load(path, encoding='utf-8', **kwargs):
    """ Return data of a json file

    Handles compression like gzip """
    if path.split(os.path.extsep)[-1] == 'gz':
        f = gzip.open(path)
        s = f.read().decode(encoding)
    else:
        f = open(path)
        s = f.read()

    data = json.loads(s)

    f.close()

    return data


@append.register(JSONLines, object)
def object_to_jsonlines(j, o, **kwargs):
    return append(j, convert(Iterator, o, **kwargs), **kwargs)


@append.register(JSONLines, Iterator)
def iterator_to_json_lines(j, seq, dshape=None, encoding='utf-8', **kwargs):
    row = next(seq)
    seq = concat([[row], seq])
    if not isinstance(row, (dict, str)) and isinstance(row, Iterable):
        seq = tuples_to_records(dshape, seq)

    lines = (json.dumps(item, default=json_dumps) for item in seq)

    # Open file
    if j.path.split(os.path.extsep)[-1] == 'gz':
        f = gzip.open(j.path, 'ab')
        lines2 = (line.encode(encoding) for line in lines)
        endl = b'\n'
    else:
        f = open(j.path, 'a')
        lines2 = lines
        endl = '\n'

    for line in lines2:
        f.write(line)
        f.write(endl)

    f.close()

    return j


@append.register(JSON, list)
def list_to_json(j, seq, dshape=None, encoding='utf-8', **kwargs):
    if not isinstance(seq[0], (dict, str)) and isinstance(seq[0], Iterable):
        seq = list(tuples_to_records(dshape, seq))
    if os.path.exists(j.path):
        with open(j.path) as f:
            if json.load(f):
                raise ValueError("Can only append to empty JSON File.\n"
                "Either remove contents from this file, save to a new file \n"
                "or use line-delimited JSON format.\n"
                "Consider using the jsonlines:// protocol, e.g.\n"
                "\todo(your-data, 'jsonlines://%s')" % j.path)

    text = json.dumps(seq, default=json_dumps)

    if j.path.split(os.path.extsep)[-1] == 'gz':
        f = gzip.open(j.path, 'wb')
        text = text.encode(encoding)
    else:
        f = open(j.path, 'w')

    f.write(text)

    f.close()
    return j


@append.register(JSON, object)
def object_to_json(j, o, **kwargs):
    return append(j, convert(list, o, **kwargs), **kwargs)


@resource.register('json://.*\.json(\.gz)?', priority=11)
def resource_json(path, **kwargs):
    if 'json://' in path:
        path = path[len('json://'):]
    return JSON(path)


@resource.register('.*\.jsonlines(\.gz)?', priority=11)
@resource.register('jsonlines://.*\.json(\.gz)?', priority=11)
def resource_jsonlines(path, **kwargs):
    if 'jsonlines://' in path:
        path = path[len('jsonlines://'):]
    return JSONLines(path)


@resource.register('.*\.json(\.gz)?')
def resource_json_ambiguous(path, **kwargs):
    """ Try to guess if this file is line-delimited or not """
    if os.path.exists(path):
        f = open(path)
        try:
            one = next(f)
        except UnicodeDecodeError:  # gzip
            f.close()
            return resource_json(path, **kwargs)
        try:
            next(f)
        except StopIteration:  # only one line
            f.close()
            return resource_json(path, **kwargs)
        try:
            json.loads(one)
            return resource_jsonlines(path, **kwargs)
        except:
            return resource_json(path, **kwargs)
        finally:
            f.close()

    # File doesn't exist, is the dshape variable length?
    dshape = kwargs.get('expected_dshape', None)
    if dshape and dshape[0] == var:
        return resource_jsonlines(path, **kwargs)
    else:
        return resource_json(path, **kwargs)


@dispatch(datetime.datetime)
def json_dumps(dt):
    s = dt.isoformat()
    if not dt.tzname():
        s = s + 'Z'
    return s


@dispatch(datetime.date)
def json_dumps(dt):
    return dt.isoformat()


@convert.register(chunks(list), (chunks(JSON), chunks(Temp(JSON))))
def convert_glob_of_jsons_into_chunks_of_lists(jsons, **kwargs):
    def _():
        return concat(convert(chunks(list), js, **kwargs) for js in jsons)
    return chunks(list)(_)


@convert.register(chunks(Iterator), (chunks(JSONLines), chunks(Temp(JSONLines))))
def convert_glob_of_jsons_into_chunks_of_lists(jsons, **kwargs):
    def _():
        return concat(convert(chunks(Iterator), js, **kwargs) for js in jsons)
    return chunks(Iterator)(_)


@convert.register(Temp(JSON), list)
def list_to_temporary_json(data, **kwargs):
    fn = '.%s.json' % uuid.uuid1()
    target = Temp(JSON)(fn)
    return append(target, data, **kwargs)


@convert.register(Temp(JSONLines), list)
def list_to_temporary_jsonlines(data, **kwargs):
    fn = '.%s.json' % uuid.uuid1()
    target = Temp(JSONLines)(fn)
    return append(target, data, **kwargs)


@drop.register((JSON, JSONLines))
def drop_json(js):
    if os.path.exists(js.path):
        os.remove(js.path)

ooc_types.add(JSONLines)



from __future__ import absolute_import, division, print_function

from contextlib import contextmanager
from warnings import warn

import networkx as nx
from datashape import discover
from .utils import expand_tuples, ignoring


ooc_types = set()  # Out-of-Core types


class FailedConversionWarning(UserWarning):
    def __init__(self, src, dest, exc):
        self.src = src
        self.dest = dest
        self.exc = exc

    def __str__(self):
        return 'Failed on %s -> %s. Working around\nError message:\n%s' % (
            self.src.__name__, self.dest.__name__, self.exc,
        )


class NetworkDispatcher(object):
    def __init__(self, name):
        self.name = name
        self.graph = nx.DiGraph()

    def register(self, a, b, cost=1.0):
        sigs = expand_tuples([a, b])

        def _(func):
            for a, b in sigs:
                self.graph.add_edge(b, a, cost=cost, func=func)
            return func
        return _

    def path(self, *args, **kwargs):
        return path(self.graph, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        return _transform(self.graph, *args, **kwargs)


def _transform(graph, target, source, excluded_edges=None, ooc_types=ooc_types,
               **kwargs):
    """ Transform source to target type using graph of transformations """
    x = source
    excluded_edges = excluded_edges or set()
    with ignoring(NotImplementedError):
        if 'dshape' not in kwargs or kwargs['dshape'] is None:
            kwargs['dshape'] = discover(x)
    pth = path(graph, type(source), target,
               excluded_edges=excluded_edges,
               ooc_types=ooc_types)
    try:
        for (A, B, f) in pth:
            x = f(x, excluded_edges=excluded_edges, **kwargs)
        return x
    except NotImplementedError as e:
        if kwargs.get('raise_on_errors'):
            raise
        warn(FailedConversionWarning(A, B, e))
        new_exclusions = excluded_edges | set([(A, B)])
        return _transform(graph, target, source, excluded_edges=new_exclusions,
                          **kwargs)


def path(graph, source, target, excluded_edges=None, ooc_types=ooc_types):
    """ Path of functions between two types """
    if not isinstance(source, type):
        source = type(source)
    if not isinstance(target, type):
        target = type(target)

    if source not in graph:
        for cls in valid_subclasses:
            if issubclass(source, cls):
                source = cls
                break

    # If both source and target are Out-Of-Core types then restrict ourselves
    # to the graph of out-of-core types
    if ooc_types:
        oocs = tuple(ooc_types)
        if issubclass(source, oocs) and issubclass(target, oocs):
            graph = graph.subgraph([n for n in graph.nodes()
                                    if issubclass(n, oocs)])
    with without_edges(graph, excluded_edges) as g:
        pth = nx.shortest_path(g, source=source, target=target, weight='cost')
        result = [(src, tgt, graph.edge[src][tgt]['func'])
                  for src, tgt in zip(pth, pth[1:])]
    return result


# Catch-all subclasses
from collections import Iterator
import numpy as np
valid_subclasses = [Iterator, np.ndarray]


@contextmanager
def without_edges(g, edges):
    edges = edges or []
    held = dict()
    for a, b in edges:
        held[(a, b)] = g.edge[a][b]
        g.remove_edge(a, b)

    try:
        yield g
    finally:
        for (a, b), kwargs in held.items():
            g.add_edge(a, b, **kwargs)



from __future__ import division, print_function, absolute_import

from datashape import var

from .. import convert, append, discover

from pyspark import SparkContext
from pyspark import RDD
from pyspark.rdd import PipelinedRDD


@append.register(SparkContext, list)
def list_to_spark_context(sc, seq, **kwargs):
    return sc.parallelize(seq)


@append.register(SparkContext, object)
def anything_to_spark_context(sc, o, **kwargs):
    return append(sc, convert(list, o, **kwargs), **kwargs)


@convert.register(list, (RDD, PipelinedRDD))
def rdd_to_list(rdd, **kwargs):
    return rdd.collect()


@discover.register(RDD)
def discover_rdd(rdd, n=50, **kwargs):
    data = rdd.take(n)
    return var * discover(data).subshape[0]



from __future__ import absolute_import, division, print_function

import functools

from toolz import merge

from multipledispatch import Dispatcher

from .convert import convert
from .append import append
from .resource import resource
from .utils import ignoring

import datashape
from datashape import discover
from datashape.dispatch import namespace
from datashape.predicates import isdimension

from .compatibility import unicode
from pandas import DataFrame, Series
from numpy import ndarray

not_appendable_types = DataFrame, Series, ndarray, tuple

__all__ = 'into',


if 'into' not in namespace:
    namespace['into'] = Dispatcher('into')

into = namespace['into']


def validate(f):
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        dshape = kwargs.pop('dshape', None)
        if isinstance(dshape, (str, unicode)):
            dshape = datashape.dshape(dshape)
        if dshape is not None and not isinstance(dshape, datashape.DataShape):
            raise TypeError('dshape argument is not an instance of DataShape')
        kwargs['dshape'] = dshape
        return f(*args, **kwargs)
    return wrapped


@into.register(type, object)
@validate
def into_type(a, b, dshape=None, **kwargs):
    with ignoring(NotImplementedError):
        if dshape is None:
            dshape = discover(b)
    return convert(a, b, dshape=dshape, **kwargs)


@into.register(object, object)
@validate
def into_object(target, source, dshape=None, **kwargs):
    """ Push one dataset into another

    Parameters
    ----------

    source: object or string
        The source of your data.  Either an object (e.g. DataFrame),
    target: object or string or type
        The target for where you want your data to go.
        Either an object, (e.g. []), a type, (e.g. list)
        or a string (e.g. 'postgresql://hostname::tablename'
    raise_on_errors: bool (optional, defaults to False)
        Raise exceptions rather than reroute around them
    **kwargs:
        keyword arguments to pass through to conversion functions.

    Examples
    --------

    >>> L = into(list, (1, 2, 3))  # Convert things into new things
    >>> L
    [1, 2, 3]

    >>> _ = into(L, (4, 5, 6))  # Append things onto existing things
    >>> L
    [1, 2, 3, 4, 5, 6]

    >>> into('myfile.csv', [('Alice', 1), ('Bob', 2)])  # doctest: +SKIP

    Explanation
    -----------

    We can specify data with a Python object like a ``list``, ``DataFrame``,
    ``sqlalchemy.Table``, ``h5py.Dataset``, etc..

    We can specify data with a string URI like ``'myfile.csv'``,
    ``'myfiles.*.json'`` or ``'sqlite:///data.db::tablename'``.  These are
    matched by regular expression.  See the ``resource`` function for more
    details on string URIs.

    We can optionally specify datatypes with the ``dshape=`` keyword, providing
    a datashape.  This allows us to be explicit about types when mismatches
    occur or when our data doesn't hold the whole picture.  See the
    ``discover`` function for more information on ``dshape``.

    >>> ds = 'var * {name: string, balance: float64}'
    >>> into('accounts.json', [('Alice', 100), ('Bob', 200)], dshape=ds)  # doctest: +SKIP

    We can optionally specify keyword arguments to pass down to relevant
    conversion functions.  For example, when converting a CSV file we might
    want to specify delimiter

    >>> into(list, 'accounts.csv', has_header=True, delimiter=';')  # doctest: +SKIP

    These keyword arguments trickle down to whatever function ``into`` uses
    convert this particular format, functions like ``pandas.read_csv``.

    See Also
    --------

    into.resource.resource  - Specify things with strings
    datashape.discover      - Get datashape of data
    into.convert.convert    - Convert things into new things
    into.append.append      - Add things onto existing things
    """
    if isinstance(source, (str, unicode)):
        source = resource(source, dshape=dshape, **kwargs)
    if type(target) in not_appendable_types:
        raise TypeError('target of %s type does not support in-place append' % type(target))
    with ignoring(NotImplementedError):
        if dshape is None:
            dshape = discover(source)
    return append(target, source, dshape=dshape, **kwargs)


@into.register((str, unicode), object)
@validate
def into_string(uri, b, dshape=None, **kwargs):
    if dshape is None:
        dshape = discover(b)

    resource_ds = 0 * dshape.subshape[0] if isdimension(dshape[0]) else dshape

    a = resource(uri, dshape=resource_ds, expected_dshape=dshape, **kwargs)
    return into(a, b, dshape=dshape, **kwargs)


@into.register((type, (str, unicode)), (str, unicode))
@validate
def into_string_string(a, b, **kwargs):
    return into(a, resource(b, **kwargs), **kwargs)


@into.register(object)
@validate
def into_curried(o, **kwargs1):
    def curried_into(other, **kwargs2):
        return into(o, other, **merge(kwargs2, kwargs1))
    return curried_into



from __future__ import absolute_import, division, print_function

import pytest

from contextlib import contextmanager
from odo.utils import tmpfile
from odo.chunks import chunks
from odo import into, append, convert, resource, discover, odo
import datashape
import pandas as pd
from datetime import datetime
import numpy as np

pytest.importorskip('tables')


df = pd.DataFrame([['a', 1, 10., datetime(2000, 1, 1)],
                   ['ab', 2, 20., datetime(2000, 2, 2)],
                   ['abc', 3, 30., datetime(2000, 3, 3)],
                   ['abcd', 4, 40., datetime(2000, 4, 4)]],
                  columns=['name', 'a', 'b', 'time'])


@contextmanager
def file(df):
    with tmpfile('.hdf5') as fn:
        f = pd.HDFStore(fn)
        f.put('/data', df, format='table', append=True)

        try:
            yield fn, f, f.get_storer('/data')
        finally:
            f.close()


def test_discover():
    with file(df) as (fn, f, dset):
        assert str(discover(dset)) == str(discover(df))
        assert str(discover(f)) == str(discover({'data': df}))


def test_discover_nested():
    with tmpfile('hdf5') as fn:
        df.to_hdf(fn, '/a/b/data')
        df.to_hdf(fn, '/a/b/data2')
        df.to_hdf(fn, '/a/data')

        hdf = pd.HDFStore(fn)

        try:
            assert discover(hdf) == discover(
                {'a': {'b': {'data': df, 'data2': df}, 'data': df}}
            )
        finally:
            hdf.close()


def eq(a, b):
    if isinstance(a, pd.DataFrame):
        a = into(np.ndarray, a)
    if isinstance(b, pd.DataFrame):
        b = into(np.ndarray, b)
    c = a == b
    if isinstance(c, np.ndarray):
        c = c.all()
    return c


def test_chunks():
    with file(df) as (fn, f, dset):
        c = convert(chunks(pd.DataFrame), dset)
        assert eq(convert(np.ndarray, c), df)


def test_resource_no_info():
    with tmpfile('.hdf5') as fn:
        r = resource('hdfstore://' + fn)
        try:
            assert isinstance(r, pd.HDFStore)
        finally:
            r.close()


def test_resource_of_dataset():
    with tmpfile('.hdf5') as fn:
        ds = datashape.dshape('{x: int32, y: 3 * int32}')
        r = resource('hdfstore://'+fn+'::/x', dshape=ds)
        try:
            assert r
        finally:
            r.parent.close()


def test_append():
    with file(df) as (fn, f, dset):
        append(dset, df)
        append(dset, df)
        assert discover(dset).shape == (len(df) * 3,)


def test_into_resource():
    with tmpfile('.hdf5') as fn:
        d = into('hdfstore://' + fn + '::/x', df)
        try:
            assert discover(d) == discover(df)
            assert eq(into(pd.DataFrame, d), df)
        finally:
            d.parent.close()


def test_convert_pandas():
    with file(df) as (fn, f, dset):
        assert eq(convert(pd.DataFrame, dset), df)


def test_convert_chunks():
    with file(df) as (fn, f, dset):
        c = convert(chunks(pd.DataFrame), dset, chunksize=len(df) / 2)
        assert len(list(c)) == 2
        assert eq(convert(pd.DataFrame, c), df)


def test_append_chunks():
    with file(df) as (fn, f, dset):
        append(dset, chunks(pd.DataFrame)([df, df]))

        assert discover(dset).shape[0] == len(df) * 3


def test_append_other():
    with tmpfile('.hdf5') as fn:
        x = into(np.ndarray, df)
        dset = into('hdfstore://'+fn+'::/data', x)
        try:
            assert discover(dset) == discover(df)
        finally:
            dset.parent.close()


def test_fixed_shape():
    with tmpfile('.hdf5') as fn:
        df.to_hdf(fn, 'foo')
        r = resource('hdfstore://'+fn+'::/foo')
        try:
            assert isinstance(r.shape, list)
            assert discover(r).shape == (len(df),)
        finally:
            r.parent.close()


def test_fixed_convert():
    with tmpfile('.hdf5') as fn:
        df.to_hdf(fn, 'foo')
        r = resource('hdfstore://'+fn+'::/foo')
        try:
            assert eq(convert(pd.DataFrame, r), df)
        finally:
            r.parent.close()


def test_append_vs_write():
    import pandas.util.testing as tm
    with tmpfile('.hdf5') as fn:
        df.to_hdf(fn, 'foo', append=True)
        store = odo(df, 'hdfstore://%s::foo' % fn)
        try:
            newdf = odo(store, pd.DataFrame)
        finally:
            store.parent.close()

    tm.assert_frame_equal(newdf, pd.concat([df, df]))

    with tmpfile('.hdf5') as fn:
        store = odo(df, 'hdfstore://%s::foo' % fn, mode='w')
        try:
            newdf = odo(store, pd.DataFrame)
        finally:
            store.parent.close()

    tm.assert_frame_equal(newdf, df)



from __future__ import print_function

import pytest

from functools import partial
import codecs
import os

from odo import odo, resource, URL, discover, CSV, TextFile, convert
from odo.backends.url import sample
from odo.temp import _Temp, Temp
from odo.utils import tmpfile, raises

import datashape

try:
    from urllib2 import urlopen
    from urllib2 import HTTPError, URLError
except ImportError:
    from urllib.request import urlopen
    from urllib.error import HTTPError, URLError

pytestmark = pytest.mark.skipif(raises(URLError,
                                       partial(urlopen, "http://google.com")),
                                reason='unable to connect to google.com')

iris_url = ('https://raw.githubusercontent.com/'
            'blaze/blaze/master/blaze/examples/data/iris.csv')
ftp_url = "ftp://athena-dist.mit.edu/pub/XNeXT/README.txt"


def test_url_resource():
    csv = resource(iris_url)
    assert isinstance(csv, URL(CSV))


def test_small_chunk_size():
    normal = convert(Temp(CSV), resource(iris_url))
    small_chunk = convert(Temp(CSV), resource(iris_url, chunk_size=1))
    with open(normal.path, 'rb') as fn:
        normal_data = fn.read()
    with open(small_chunk.path, 'rb') as fn:
        small_chunk_data = fn.read()
    assert normal_data == small_chunk_data



def test_sample_different_line_counts():
    with sample(resource(iris_url), lines=10) as fn:
        with open(fn, 'r') as f:
            assert len(list(f)) == 10

    with sample(resource(iris_url), lines=5) as fn:
        with open(fn, 'r') as f:
            assert len(list(f)) == 5


def test_sample_different_encoding():
    encoding = 'latin-1'
    lines = 10
    with sample(resource(iris_url), lines=lines, encoding=encoding) as fn:
        with codecs.open(fn, 'r', encoding=encoding) as f:
            assert len(list(f)) == lines


@pytest.mark.xfail(raises=HTTPError)
def test_failed_url():
    failed_url = "http://foo.com/myfile.csv"
    with tmpfile('.csv') as fn:
        odo(failed_url, fn)


def test_url_discover():
    csv = resource(iris_url)
    assert isinstance(discover(csv), datashape.DataShape)


def test_url_to_local_csv():
    with tmpfile('.csv') as fn:
        csv = odo(iris_url, fn)
        path = os.path.abspath(csv.path)
        assert os.path.exists(path)


def test_url_txt_resource():
    txt = resource(ftp_url)
    assert isinstance(txt, URL(TextFile))


@pytest.mark.xfail(
    raises=URLError,
    reason='MIT Athena FTP is down as of October 23, 2015'
)
def test_ftp_to_local_txt():
    with tmpfile('.txt') as fn:
        txt = odo(ftp_url, fn, timeout=5)
        path = os.path.abspath(txt.path)
        assert os.path.exists(path)


def test_convert():
    url_csv = resource(iris_url)
    t_csv = convert(Temp(CSV), url_csv)
    assert discover(url_csv) == discover(t_csv)

    assert isinstance(t_csv, _Temp)


@pytest.mark.skipif(os.environ.get('HDFS_TEST_HOST') is None,
                    reason='No HDFS_TEST_HOST envar defined')
def test_url_to_hdfs():
    from .test_hdfs import tmpfile_hdfs, hdfs, HDFS

    with tmpfile_hdfs() as target:

        # build temp csv for assertion check
        url_csv = resource(iris_url)
        csv = convert(Temp(CSV), url_csv)

        # test against url
        scsv = HDFS(CSV)(target, hdfs=hdfs)
        odo(iris_url, scsv)

        assert discover(scsv) == discover(csv)
